Struct that represents a combination of expressions that all must be true.
An And without children evaluates to true.
Summary
Types
@type t() :: %Spek.AllOf{children: [Spek.expression()], satisfied?: boolean() | nil}
Struct that represents a combination of expressions that all must be true.
An And without children evaluates to true.
@type t() :: %Spek.AllOf{children: [Spek.expression()], satisfied?: boolean() | nil}