Struct that represents a combination of checks one of which must be true.
An Or without children evaluates to false.
Summary
Types
@type t() :: %Spek.AnyOf{children: [Spek.expression()], satisfied?: boolean() | nil}
Struct that represents a combination of checks one of which must be true.
An Or without children evaluates to false.
@type t() :: %Spek.AnyOf{children: [Spek.expression()], satisfied?: boolean() | nil}