Struct that represents an authorization rule that evaluates to a fixed value.
Summary
Types
@type t() :: %Spek.Literal{result: Spek.truthy(), satisfied?: true} | %Spek.Literal{result: Spek.falsy(), satisfied?: false}
Struct that represents an authorization rule that evaluates to a fixed value.
@type t() :: %Spek.Literal{result: Spek.truthy(), satisfied?: true} | %Spek.Literal{result: Spek.falsy(), satisfied?: false}